import typing as typ

import bpy
import pydantic as pyd


class BLIDProps:
	"""Set of properties, defined by `BLField`, which should be registered to an existing `bpy.types.ID` class."""

	bl_id_type: bpy.types.ID

	@classmethod
	def bl_fields(cls) -> list[typ.Any]:
		"""Return BLFields defined on this class."""
		# Return BLField Entries
		## - It would be better to check isinstance(..., BLField).
		## - However, this would create a circular reference.
		return [
			attr_value
			for attr_name, attr_value in vars(cls).items()
			if hasattr(attr_value, 'bl_prop')
			and hasattr(attr_value, 'init_on_bl_class')
		]

	@classmethod
	def register(cls) -> None:
		"""Registers all fields to the given ID class."""
		if hasattr(cls, 'bl_fields'):
			for bl_field in cls.bl_fields():
				setattr(cls.bl_id_type, bl_field.bl_prop.name, bl_field)
				bl_field.bl_prop.init_on_bl_class(cls.bl_id_type)
				bl_field.init_on_bl_class(cls.bl_id_type)

			# IDEA: Nice spot to generate a bunch of bl_instance.instance_id's?

		else:
			msg = 'No BLFields are defined on this class.'
			raise ValueError(msg)

	@classmethod
	def unregister(cls) -> None:
		"""Unregisters all fields from the given ID class."""
		pass
		## TODO: Implement?
