# blroots
# Copyright (C) 2025 blroots Project Contributors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Implements `IntProp`."""

import builtins
import functools
import typing as typ

import bpy

from ..bl_prop import BLProp, SupportedBLPropType

PRIORITY = 10


class IntProp(BLProp, frozen=True):
	"""A single constrained integer number.

	Parameters:
		abs_min: The absolute minimum value of the inting point number.
		abs_max: The absolute maximum value of the inting point number.
		soft_min: The lowest value that can be set using the UI slider.
		soft_max: The highest value that can be set using the UI slider.
	"""

	default_value: int

	# Constraints
	abs_min: int | None = None
	abs_max: int | None = None

	# UI Elements
	soft_min: int | None = None
	soft_max: int | None = None

	####################
	# - Type Support
	####################
	@classmethod
	def supports_type_hint(cls, field_type_hint: type) -> bool:
		"""Support only the explicit type `int` aka. `builtins.int`."""
		return field_type_hint is builtins.int

	####################
	# - bpy.props Interface
	####################
	@property
	def bpy_prop_type(self) -> type[bpy.props.IntProperty]:
		"""The underlying `bpy.props` class backing this property."""
		return bpy.props.IntProperty  # type: ignore[no-any-return]

	@functools.cached_property
	def bpy_prop_kwargs(self) -> dict[str, typ.Any]:
		"""Keyword arguments to pass to `self.bpy_prop_type`."""
		kwargs: dict[str, typ.Any] = {
			'default': self.default_value,
		}

		if self.abs_min is not None:
			kwargs |= {'min': self.abs_min}
		if self.abs_max is not None:
			kwargs |= {'max': self.abs_max}
		if self.soft_min is not None:
			kwargs |= {'soft_min': self.soft_min}
		if self.soft_max is not None:
			kwargs |= {'soft_min': self.soft_max}

		return kwargs

	####################
	# - Encode/Decode
	####################
	def encode(self, value: typ.Any) -> int:
		"""Encode a value that supports `int()` to a Blender int."""
		# Parse the Value
		try:
			parsed_value = int(value)
		except ValueError as ex:
			msg = f"Value '{value}' could not be coerced to a int."
			raise ValueError(msg) from ex

		# Snap Value to Bounds
		if self.abs_min is not None and parsed_value < self.abs_min:
			return self.abs_min
		if self.abs_max is not None and parsed_value > self.abs_max:
			return self.abs_max
		return parsed_value

	def decode(self, raw_bl_value: int) -> int:
		"""Decode a Blender int by doing nothing."""
		return raw_bl_value

	####################
	# - UI
	####################
	def draw(
		self,
		bl_instance: SupportedBLPropType,
		layout: bpy.types.UILayout,
		slider: bool = False,
	) -> None:
		"""Draw this property to the given Blender `UILayout`."""
		## TODO:
		## - Support for slider
		layout.prop(bl_instance, self.bl_name, text=self.display_name, slider=slider)

	####################
	# - Selection
	####################
	@classmethod
	def priority(cls) -> int:
		"""Priority with which to prefer this implementation over others compatible with the same type hint."""
		return PRIORITY
