# blroots
# Copyright (C) 2025 blroots Project Contributors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Tools that gather extension components from submodules."""

import functools
import sys
import tomllib
import types
from pathlib import Path

import bpy
import pydantic as pyd

from blroots import typ as bltyp
from blroots.utils import bl_logger
from blroots.utils.bl_handlers import BLHandlers
from blroots.utils.bl_keymap_item import BLKeymapItem


####################
# - Types
####################
class BLExtInfo(pyd.BaseModel):
	"""Information about a currently loaded Blender extension."""

	# Manifest Fields
	name: bltyp.BLExtName
	pkg_name: bltyp.BLExtPkgName

	version: bltyp.BLExtVersion

	# Init Settings
	use_log_file: bool
	log_file_path: Path
	log_file_level: bl_logger.BLExtLogLevel

	use_log_console: bool
	log_console_level: bl_logger.BLExtLogLevel

	def operator_name(self, operator_name_base: str) -> str:
		"""Conform a string to a correct `bpy.types.Operator` name for this extension.

		In particular, operator names must have the format `<addon_name>.<...>`.
		"""
		return f'{self.name}.{operator_name_base}'

	def panel_name(self, panel_name_base: str) -> str:
		"""Conform a string to a correct `bpy.types.Panel` name for this extension.

		In particular, panel names must have the format `<ADDON_NAME_UPPERCASE>_PT_<...>`.
		"""
		return f'{self.name.upper()}_PT_{panel_name_base}'


####################
# - Gather Classes
####################
def gather_classes(*modules: types.ModuleType) -> list[type[bltyp.BLClass]]:
	"""Gather any classes present in the `BL_REGISTER` attribute of the passed modules."""
	# Validate Presence of BL_REGISTER
	for module in modules:
		if not hasattr(module, 'BL_REGISTER'):
			msg = f"The module {module.__name__} has no 'BL_REGISTER' attribute, so no Blender classes could be gathered from it."
			raise ValueError(msg)

	# Gather BLClasses from Modules
	return [bl_class for module in modules for bl_class in module.BL_REGISTER]


classes = gather_classes


####################
# - Gather Handlers
####################
def gather_handlers(*modules: types.ModuleType) -> BLHandlers:
	"""Gather any classes present in the `BL_HANDLERS` attribute of the passed modules."""
	# Validate Presence of BL_REGISTER
	for module in modules:
		if not hasattr(module, 'BL_HANDLERS'):
			msg = f"The module {module.__name__} has no 'BL_HANDLERS' attribute, so no Blender handlers could be gathered from it."
			raise ValueError(msg)

	# Gather Handlers from Modules
	return functools.reduce(
		lambda a, b: a + b,
		[
			module.BL_HANDLERS if module.BL_HANDLERS is not None else BLHandlers()
			for module in modules
		],
		BLHandlers(),
	)


handlers = gather_handlers


####################
# - Gather Keymap Items
####################
def gather_keymap_items(*modules: types.ModuleType) -> list[BLKeymapItem]:
	"""Gather any classes present in the `BL_KEYMAP_ITEMS` attribute of the passed modules."""
	# Validate Presence of BL_REGISTER
	for module in modules:
		if not hasattr(module, 'BL_KEYMAP_ITEMS'):
			msg = f"The module {module.__name__} has no 'BL_KEYMAP_ITEMS' attribute, so no Blender keymap items could be gathered from it."
			raise ValueError(msg)

	# Gather Keymap Items from Modules
	return [
		bl_keymap_item
		for module in modules
		for bl_keymap_item in (
			module.BL_KEYMAP_ITEMS if module.BL_KEYMAP_ITEMS is not None else []
		)
	]


keymap_items = gather_keymap_items


####################
# - Gather Addon Information
####################
def gather_path_blext_root(blext_pkg: str) -> Path:
	"""Deduce the top-level path to the installed Blender extension.

	Uses the `__file__` attribute of its `sys.modules` entry.
	"""
	blext_module = sys.modules.get(blext_pkg)
	if blext_module is not None:
		if blext_module.__file__ is not None:
			return Path(blext_module.__file__).resolve().parent

		msg = f"Can't resolve path to '{blext_pkg}', since its 'sys.modules' entry has no '__file__' attribute."
		raise RuntimeError(msg)
	msg = f"Can't find module for '{blext_pkg}' in 'sys.modules'."
	raise ValueError(msg)


####################
# - Gather Addon Information
####################
@functools.cache
def gather_blext_info(
	blext_pkg: bltyp.BLExtPkgName,
) -> BLExtInfo:
	"""Parse information about a running extension from its manifest and package name."""
	path_blext_root = gather_path_blext_root(blext_pkg)
	path_manifest: Path = path_blext_root / 'blender_manifest.toml'
	path_init_settings: Path = path_blext_root / 'init_settings.toml'

	with path_manifest.open('rb') as f:
		manifest = tomllib.load(f)

	with path_init_settings.open('rb') as f:
		init_settings = tomllib.load(f)

	writable_blext_dir = Path(
		bpy.utils.extension_path_user(
			blext_pkg,
			path='',
			create=True,
		)
	)
	return BLExtInfo(
		name=manifest['id'],
		pkg_name=blext_pkg,
		version=manifest['version'],
		use_log_file=init_settings['use_log_file'],
		log_file_path=writable_blext_dir / init_settings['log_file_path'],
		log_file_level=init_settings['log_file_level'],
		use_log_console=init_settings['use_log_console'],
		log_console_level=init_settings['log_console_level'],
	)


## TODO: Clear the cache entry of an unregister()ing extension.


def blext_info(module_name: str) -> BLExtInfo:
	"""Get the `BLExtInfo` from a `__name__` module path."""
	blext_pkg = '.'.join(module_name.split('.')[:3])
	return gather_blext_info(blext_pkg)
