# blext
# Copyright (C) 2025 blext Project Contributors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""A lightweight, `rich`-based approach to logging that is isolated from other extensions that may used the Python stdlib `logging` module."""

import enum
import functools
import logging
import typing as typ
from pathlib import Path

import bpy
import rich.console
import rich.logging
import rich.traceback

from blroots import gather

####################
# - Configuration
####################
STREAM_LOG_FORMAT = 11 * ' ' + '%(levelname)-8s %(message)s (%(name)s)'
FILE_LOG_FORMAT = STREAM_LOG_FORMAT

OUTPUT_CONSOLE = rich.console.Console(
	color_system='truecolor',
)
ERROR_CONSOLE = rich.console.Console(
	color_system='truecolor',
	stderr=True,
)


####################
# - Types
####################
LogLevel: typ.TypeAlias = int


class BLExtLogLevel(enum.StrEnum):
	"""Log level of a Blender extension.

	Attributes:
		Debug: Low-level messages for debugging.
		Info: Communications about routine operations.
		Warning: Something that might be wrong.
		Error: Something wrong that can be recovered from.
		Critical: Something wrong that cannot be recovered from.
	"""

	Debug = enum.auto()
	Info = enum.auto()
	Warning = enum.auto()
	Error = enum.auto()
	Critical = enum.auto()

	@property
	def bl_display_name(self) -> str:
		"""Convert the enum value to a human-friendly name.

		Returns:
			A human-friendly name corresponding to the enum value.
		"""
		LL = BLExtLogLevel
		return {
			LL.Debug: 'Debug',
			LL.Info: 'Info',
			LL.Warning: 'Warning',
			LL.Error: 'Error',
			LL.Critical: 'Critical',
		}[self]

	@property
	def bl_icon(self) -> str:
		"""Convert the enum value to a Blender icon.

		Returns:
			A human-friendly name corresponding to the enum value.
		"""
		return ''

	@property
	def log_level(self) -> int:
		"""Standard corresponding log-level integer from the `logging` module."""
		LL = BLExtLogLevel
		return {
			LL.Debug: logging.DEBUG,
			LL.Info: logging.INFO,
			LL.Warning: logging.WARNING,
			LL.Error: logging.ERROR,
			LL.Critical: logging.CRITICAL,
		}[self]


####################
# - Naming
####################
@functools.cache
def blext_logger_name(blext_pkg: str) -> str:
	"""Name of the root logger for a particular Blender extension name."""
	return f'blext__{blext_pkg}'


####################
# - Logger Access
####################
def all_blext_loggers(blext_pkg: str) -> frozenset[logging.Logger]:
	"""Retrieve all loggers currently declared by this addon.

	These loggers are all children of a logger defined by `blext_logger_name`.
	This allows for name-isolation from other Blender extensions, as well as easy cleanup.

	Returns:
		Set of all loggers declared by this addon.
	"""
	return frozenset(
		{
			logging.getLogger(name)
			for name in logging.root.manager.loggerDict
			if name.startswith(blext_logger_name(blext_pkg))
		}
	)
	## TODO: Python 3.12 has a .getChildren() method that also returns sets.


####################
# - Logger Setup
####################
def get(module_name: str) -> logging.Logger:
	"""Retrieve and/or create a logger corresponding to a module name.

	Examples:
		Should generally be used as:
		```python
		log = logger.get(__name__)
		```

		The `log` variable can be considered a shorthand for `logger`, since it is a `logging.Logger` object.

	Parameters:
		blext_pkg: The `__package__` variable of the top-level extension package.
		module_name: The `__name__` variable of the module to return a logger for.
	"""
	if module_name.startswith('blroots'):
		return logging.getLogger('blroots')

	# Deduce BLExt Package Name from Module Name
	blext_pkg = gather.blext_info(module_name).pkg_name

	# Retrieve / Create Child Logger
	addon_root_logger: logging.Logger = logging.getLogger(blext_logger_name(blext_pkg))
	logger = addon_root_logger.getChild(module_name)

	# Addon Prefs Integration: Conform Logger to Addon Preferences
	## - Requires the 'bl_logger' integration to be enabled.
	addon = bpy.context.preferences.addons.get(blext_pkg)
	if (
		addon is not None
		and (addon_prefs := addon.preferences) is not None
		and getattr(addon_prefs, 'enable_integration__bl_logger', False)
	):
		addon_prefs.conform_logger(logger)

	# Conform Logger to Init Settings
	else:
		blext_info = gather.gather_blext_info(blext_pkg)
		update_logger(
			logger,
			cb_console_handler=sensible_console_handler,
			cb_file_handler=sensible_file_handler,
			console_level=(
				blext_info.log_console_level.log_level
				if blext_info.use_log_console
				else None
			),
			file_path=blext_info.log_file_path if blext_info.use_log_file else None,
			file_level=blext_info.log_file_level.log_level
			if blext_info.use_log_file
			else None,
		)

	return logger


####################
# - Logger Update
####################
def update_logger(
	logger: logging.Logger,
	*,
	cb_console_handler: typ.Callable[[LogLevel], logging.Handler],
	cb_file_handler: typ.Callable[[Path, LogLevel], logging.Handler],
	console_level: LogLevel | None,
	file_path: Path | None,
	file_level: LogLevel | None,
) -> None:
	"""Configures a single logger with given console and file handlers, individualizing the log level that triggers it.

	This is a lower-level function - generally, modules that want to use a well-configured logger will use the `get()` function, which retrieves the parameters for this function from the addon preferences.
	This function is used by the higher-level log setup.

	Parameters:
		cb_console_handler: A function that takes a log level threshold (inclusive), and returns a logging handler to a console-printer.
		cb_file_handler: A function that takes a log level threshold (inclusive), and returns a logging handler to a file-printer.
		logger: The logger to configure.
		console_level: The log level threshold to print to the console.
			None deactivates file logging.
		path_log_file: The path to the log file.
			None deactivates file logging.
		file_level: The log level threshold to print to the log file.
	"""
	####################
	# - Initialize Logger
	####################
	# DO NOT Propagate to Root Logger
	## - This looks like 'double messages'
	## - See SO/6729268/log-messages-appearing-twice-with-python-logging
	logger.propagate = False

	# Let All Messages Through
	## - The individual handlers perform appropriate filtering.
	logger.setLevel(logging.NOTSET)

	# Clear All Existing Handlers
	if logger.handlers:
		logger.handlers.clear()

	####################
	# - Add Logger Handlers
	####################
	# Add Console Logging Handler
	if console_level is not None:
		logger.addHandler(cb_console_handler(console_level))

	# Add File Logging Handler
	if file_path is not None and file_level is not None:
		logger.addHandler(cb_file_handler(file_path, file_level))


####################
# - Default Logging Handlers
####################
def sensible_console_handler(level: LogLevel) -> rich.logging.RichHandler:
	"""A logging handler that prints messages to the console.

	Parameters:
		level: The log levels (debug, info, etc.) to print.

	Returns:
		The logging handler, which can be added to a logger.
	"""
	rich_formatter = logging.Formatter(
		'%(message)s',
		datefmt='[%X]',
	)
	rich_handler = rich.logging.RichHandler(
		level=level,
		console=ERROR_CONSOLE,
		rich_tracebacks=True,
	)
	rich_handler.setFormatter(rich_formatter)
	return rich_handler


def sensible_file_handler(
	path_log_file: Path, level: LogLevel
) -> rich.logging.RichHandler:
	"""A logging handler that prints messages to a file.

	Parameters:
		path_log_file: The path to the log file.
		level: The log levels (debug, info, etc.) to append to the file.

	Returns:
		The logging handler, which can be added to a logger.
	"""
	file_formatter = logging.Formatter(FILE_LOG_FORMAT)
	file_handler = logging.FileHandler(path_log_file)
	file_handler.setFormatter(file_formatter)
	file_handler.setLevel(level)
	return file_handler
